using System;
using System.Configuration;
using System.Diagnostics;
using System.Web;
using System.Web.SessionState;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;

namespace ReverseProxy {
	class HandlerFactory : IHttpHandlerFactory {
		public IHttpHandler GetHandler(HttpContext context, string requestType, String url, String pathTranslated) {
			if(url.ToLower().EndsWith("logon.aspx"))
				return System.Web.UI.PageParser.GetCompiledPageInstance(url, pathTranslated, context);
      else
				return new ReverseProxy();
		}

		public void ReleaseHandler(IHttpHandler handler) { }

		public bool IsReusable {
			get {return true;}
    }
  }

  public class ReverseProxy: IHttpHandler, IRequiresSessionState {
  	public HttpWebRequest createProxyRequest(HttpRequest originalRequest, Uri uri) {
      HttpWebRequest proxyRequest = (HttpWebRequest)WebRequest.Create(uri);
      proxyRequest.Method = originalRequest.HttpMethod;
      proxyRequest.ContentType = originalRequest.ContentType;
      proxyRequest.UserAgent = originalRequest.UserAgent;

			foreach(String each in originalRequest.Headers) {
				if (!WebHeaderCollection.IsRestricted(each))
					proxyRequest.Headers.Add(each, originalRequest.Headers.Get(each));
			}

      if (originalRequest.ContentLength > 0) {
        Stream outputStream = proxyRequest.GetRequestStream();
      	pipeStream(originalRequest.InputStream, outputStream);
        outputStream.Close();
      }

      proxyRequest.AllowAutoRedirect = false;

			if (HttpContext.Current.Session != null && HttpContext.Current.Session.Count == 2) {
				proxyRequest.PreAuthenticate = true;
      	proxyRequest.Credentials = new NetworkCredential(HttpContext.Current.Session["userid"].ToString(), HttpContext.Current.Session["passwd"].ToString());
			}
      return proxyRequest;
  	}

		private bool isRedirection(HttpStatusCode code) {
			string statusCode = Enum.Format(typeof(HttpStatusCode), code, "d");

			return statusCode.StartsWith("3");
		}

    public bool IsReusable {
      get {return true;}
    }

    public Uri mapUrl(string appPath, Uri url) {
      string remoteWebSite = ConfigurationSettings.AppSettings["RemoteWebSite"];
      return new Uri(url.AbsoluteUri.Replace(url.Scheme + Uri.SchemeDelimiter + url.Authority + appPath, remoteWebSite));
    }

    private string parseHtmlResponse(string html, string appPath) {
    	/*
				substitute in the application path to fix up urls like:

    		href="/<proxiedUrl>"	->	href="/<appPath>/<proxiedUrl>"
        href='/<proxiedUrl>'	->	href="/<appPath>/<proxiedUrl>"		(handle single quoted urls)

				also works with "src" and any other token to the left of the equal sign other than "value" which are possibly html inputs on screen
			*/

    	html = Regex.Replace(html, "(?<!value=)(?:\"|')/(\\S*?)(?:\"|')", "\"" + appPath + "/" + "$1" + "\"", RegexOptions.Multiline);

      html = html.Replace("=/","=" + appPath + "/");
      return html;
    }

 		private string parseRealm(string authHeader) {
			Regex regex = new Regex(".*=\\\"(.*)\"");

			Match match = regex.Match(authHeader);
			if (match.Success)
				return match.Groups[1].Value;
			else
				return "";
		}

    private void pipeStream(Stream input, Stream output) {
			Byte[] buff = new byte[1024];
			int bytes = 0;

			while( (bytes = input.Read(buff, 0, 1024) ) > 0 )
				output.Write(buff, 0, bytes);
    }

    public void ProcessRequest(HttpContext context) {
			Uri url = context.Request.Url;
			Uri remoteUrl = mapUrl(context.Request.ApplicationPath, url);

      HttpWebRequest proxyRequest;
      HttpWebResponse response;
      try {
      	do {
		      proxyRequest = createProxyRequest(context.Request, remoteUrl);

  	      response = (HttpWebResponse)proxyRequest.GetResponse();
    	    if (isRedirection(response.StatusCode))
    	    	remoteUrl = new Uri(remoteUrl.Scheme + Uri.SchemeDelimiter + remoteUrl.Authority + response.Headers["Location"]);

      	} while (isRedirection(response.StatusCode));
      }
      catch(System.Net.WebException e) {
      	HttpWebResponse webResponse = e.Response as HttpWebResponse;
      	if (webResponse != null) {
      		if (webResponse.StatusCode == HttpStatusCode.Unauthorized) {
      			string realm = parseRealm(webResponse.GetResponseHeader("WWW-AUTHENTICATE"));
	      		context.Response.Redirect(context.Request.ApplicationPath + "/logon.aspx?Realm=" + context.Server.UrlEncode(realm) + "&ReturnUrl=" + context.Server.UrlEncode(url.PathAndQuery));
      			return;
      		}

	        context.Response.StatusCode = (int)webResponse.StatusCode;
	        context.Response.StatusDescription = webResponse.StatusDescription;
				}

				if (e.Response == null) {
					context.Response.Write("<p>" + e.Status + "</p>");
					context.Response.Write("<p>" + e.Message + "</p>");
				}
				else {
	      	context.Response.ContentType = e.Response.ContentType;
  	      Stream responseStream = e.Response.GetResponseStream();
    	  	pipeStream(responseStream, context.Response.OutputStream);
      	  responseStream.Close();
				}

        context.Response.End();
        return;
      }

      processResponse(context, response);
		}

		private void processResponse(HttpContext context, HttpWebResponse response) {
      try {
	      Stream receiveStream = response.GetResponseStream();
	      context.Response.ContentType = response.ContentType;

	      try {
					if ((response.ContentType.ToLower().IndexOf("html") >= 0) || (response.ContentType.ToLower().IndexOf("javascript")>=0)) {
		        StreamReader readStream = new StreamReader(receiveStream, Encoding.Default);
		        string content;

		        content = parseHtmlResponse(readStream.ReadToEnd(), context.Request.ApplicationPath);

		        context.Response.Write(content);
		        readStream.Close();
		      }
		      else
		      	pipeStream(receiveStream, context.Response.OutputStream);
		    }
				finally {
					context.Response.End();
				}
      }
      finally {
	      response.Close();
	    }
    }
  }
}
